/*  
    VTun - Virtual Tunnel over TCP/IP network.

    Copyright (C) 1998-2000  Maxim Krasnyansky <max_mk@yahoo.com>

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 */

/*
 * VTun additions for SecoNET gateways.
 *
 * Schrack/TUV183-1 GG 04/2015
 *   Generic functions for AES en/decryption.
 *   Used in auth.c (ECB) and lfd_encrypt.c (GCM)
 */ 

/*
   * AES and RNG implementation from the polarSSL kit:
   * https://tls.mbed.org/
   * Copyright (C) 2006-2012, Brainspark B.V. All rights reserved.
   * Lead Maintainer: Paul Bakker <polarssl_maintainer at polarssl.org>
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <time.h>

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#include "vtun.h"
#include "linkfd.h"

/* Encryption and Decryption of the challenge key */
#ifdef HAVE_SSL

#include "Schrack_AES.h"

// Error handling
static int f_lastError = 0;

// Crypto key and contexts
static char f_keyAES[ AESGCM_KEY_LENGTH ];
	// AES key, MD5-hashed from the VTUN password in the init functions
static TCtxAES f_ctxAES;
	// AES context, initialized in AES_Init()
static TCtxGCM f_ctxGCM;
	// AES-GCM context, initialized in AESGCM_Init()


/* Generic helper functions ***************************************************/

int AES_GetError( void ) { return f_lastError; }

	// Generate a 128 bit key by hashing the password
#if !defined(AES_BLOCKSIZE_BITS) || AES_BLOCKSIZE_BITS != 128
#error "Key generation needs to be adapted for block size"
#endif
static inline void HashKey( char* key, int keyLength )
{
	md5( (uint8*)key, keyLength, (uint8*)f_keyAES );
}

/* Random number generation ***************************************************/

	// Initialize RNG context with an arbitrary personalization string
	// (and possibly add entropy sources).
	// Used in auth.c to generate a random challenge for authentification.
static inline int Rnd_Init( ctr_drbg_context* pRndContext )
{
	char* ps = "sR4\"z\r'ß=<\n"; // Arbitrary string for personalization
		// NOTE: possibly add in password/interface/ip
	entropy_context ec;
		// TODO: Check if entropy sources are available

	entropy_init( &ec );
	
	f_lastError = ctr_drbg_init( pRndContext, entropy_func, &ec, (unsigned char *) ps, strlen( ps ) );
	
	//entropy_free( &ec ); // Only used in later versions of polarSSL

	return f_lastError;
}


	// Generate a random byte sequence.
	// NOTE: Perhaps initialize only one drbg context for the module?
int RAND_bytes( uint8* bytes, int count )
{
   ctr_drbg_context rndCtx;

   f_lastError = Rnd_Init( &rndCtx );
   if( f_lastError != 0 ) return f_lastError;

   f_lastError = ctr_drbg_random( &rndCtx, bytes, count );

   //ctr_drbg_free( &rndCtx ); // Only used in later versions of polarSSL

   return f_lastError;
}


/* AES en/decryption **********************************************************/

	// Initialize the AES module for encryption.
int AES_InitEncrypt( char* key, int keyLength )
{
	HashKey( key, keyLength );
		// NOTE: Should we risk storing the key for the lifetime of the app?
	f_lastError = aes_setkey_enc( &f_ctxAES, (uint8*)f_keyAES, AESGCM_KEYSIZE_BITS);
	memset( f_keyAES, 0, AES_KEY_LENGTH );

	return f_lastError;
}

	// Initialize for decryption.
	// keyLength is given in bits (must be one of 128, 192, or 256)
int AES_InitDecrypt( char* key, int keyLength )
{
	HashKey( key, keyLength );
	f_lastError = aes_setkey_dec( &f_ctxAES, (uint8*)f_keyAES, AESGCM_KEYSIZE_BITS );
	memset( f_keyAES, 0, AES_KEY_LENGTH );
	return f_lastError;
}


// Cleanup of AES module.
void AES_Exit( void )
{
	//memset( f_aeskey, 0, AESGCM_KEY_LENGTH );
}


// Encrypt a block of 16 bytes with 128 bit AES-ECB
int AES_EncryptECB( char* aInput, char* aOutput )
{
	f_lastError = aes_crypt_ecb( &f_ctxAES, AES_ENCRYPT, (uint8*)aInput, (uint8*)aOutput );
	return f_lastError != 0 ? 0 : 16; // On error, return zero length
}

// Decrypt a block of 16 bytes with 128 bit AES-ECB
int AES_DecryptECB( char* aInput, char* aOutput )
{
	f_lastError = aes_crypt_ecb( &f_ctxAES, AES_DECRYPT, (uint8*)aInput, (uint8*)aOutput );
	return f_lastError != 0 ? 0 : 16; // On error, return zero length
}


/* AES-GCM en/decryption ******************************************************/

// Initialize the GCM module.
// keysize is given in bits (must be one of 128, 192, or 256)
int AESGCM_Init( char* key, int keyLength )
{
	HashKey( key, keyLength );
	f_lastError = gcm_init( &f_ctxGCM, (uint8*)f_keyAES, AESGCM_KEYSIZE_BITS );
	memset( f_keyAES, 0, AESGCM_KEY_LENGTH );
	return f_lastError;
}

// Cleanup of GCM module.
void AESGCM_Exit( void )
{
	//memset( f_aeskey, 0, AESGCM_KEY_LENGTH );
}

// Helper functions to retrieve pointers to various elements in the header.

static inline size_t LengthAAD( void ) { return AESGCM_AAD_LENGTH; }
static inline size_t LengthTAG( void ) { return AESGCM_TAG_LENGTH; }
static inline size_t LengthIV( void ) { return AESGCM_IV_LENGTH; }
static inline size_t LengthTrail( void ) { return AESGCM_BUFFER_TRAIL; }

static inline size_t NetLength( int length ) { return length - AESGCM_HEADER_LENGTH; }
   // Net length of packet excluding the authentication header

   // The payload VTUN hands to us consists of our encryption header followed by
   // the tunneled packet. Header consists of AAD, TAG, and IV.
static inline uint8* BufferAAD( char* buffer ) {
   return (uint8*)buffer; }
static inline uint8* BufferTAG( char* buffer ) {
   return BufferAAD( buffer ) + LengthAAD( ); }
static inline uint8* BufferIV( char* buffer ) { // NOTE: Use 12 payload bytes as IV?
   return BufferTAG( buffer ) + LengthTAG( ); }
static inline uint8* BufferInput( char* buffer ) {
   return BufferIV( buffer ) + LengthIV( ); }

   // Output buffer is in different locations for encoding and decoding:
   // When encoding, the input buffer can be re-used (see gcm.h)
static inline uint8* BufferEncOutput( char* buffer ) {
   return BufferInput( buffer ); }

   // When decoding, the IV is copied out beforehand, and part of the space is
   // used for the output trailing offset of 8 bytes.
static inline uint8* BufferDecOutput( char* buffer ) {
   return BufferInput( buffer ) - LengthTrail( ); }


static inline TAddData* AAD_Pointer( char* buffer ) {
   return (TAddData*)BufferAAD( buffer ); }

static inline void AAD_SetSeqNumber( char* buffer, uint16 num ) {
   AAD_Pointer( buffer )->seqNum = num; }

   // Encrypt the data in the given buffer (in place) with the given AES key
   // key has to be of appropriate size (128/192/256 bit)
   // buffer holds AES header, VTUN header, payload
   // length is the length of the total message including headers, in bytes
int AESGCM_Encrypt( char* buffer, int length, char** pOutput )
{
   static uint16 s_seqNum = 0;
   // Create a random initialization vector and put it into the message buffer,
   // so it is transmitted to the receiver along with the data
   f_lastError = RAND_bytes( BufferIV( buffer ), LengthIV( ) );
   if( f_lastError != 0 ) return f_lastError;

   AAD_SetSeqNumber( buffer, s_seqNum++ ); // Set the current sequence number
   *pOutput = buffer; // The entire buffer including overhead needs to be transmitted

	f_lastError = gcm_crypt_and_tag(
		&f_ctxGCM, GCM_ENCRYPT, NetLength( length ), // length of input to be encrypted
		BufferIV( buffer ), LengthIV( ), // initialization vector and length (in bytes),
		BufferAAD( buffer ), LengthAAD( ), // additional data and length (bytes)
		BufferInput( buffer ),   // input pointer
		BufferEncOutput( buffer ),  // output pointer
		LengthTAG( ), BufferTAG( buffer ) ); // tag and length (bytes)

	// On error, return zero length; otherwise, entire length is used, since the
	// AES header is sent along with the payload.
	return f_lastError != 0 ? 0 : length;
}

	// Decrypt the data in the given buffer with the given AES key. We use only
	// one buffer, so the output pointer must be offset by 8 bytes ahead of the
	// input (see comments in gcm.h); BufferOutput() takes care of that.
	// Returns the length of the decrypted buffer (excluding the header)
	// buffer holds AES header and payload.
	// length is the length of the total message including header, in bytes.
	// pDecoded will be set to point to the output buffer (the actual packet)
	//TODO: Check authentication tag
int AESGCM_Decrypt( char* buffer, int length, char** pDecoded )
{
	char initVect[ AESGCM_IV_LENGTH ];

	// Read initialization vector from message (space is re-used for buffer trailing)
	memcpy( initVect, BufferIV( buffer ), AESGCM_IV_LENGTH );

	*pDecoded = (char*)BufferDecOutput( buffer ); // after decoding, we only use the payload

	f_lastError = gcm_crypt_and_tag(
		&f_ctxGCM, GCM_DECRYPT, NetLength( length ), // context, mode=decrypt, length of input buffer
		BufferIV( buffer ), LengthIV( ), // initialization vector and length (in bits),
		BufferAAD( buffer ), LengthAAD( ), // additional data and length
		BufferInput( buffer ),   // input pointer
		BufferDecOutput( buffer ),  // output pointer
		LengthTAG( ), BufferTAG( buffer ) ); // tag and length

	// Again zero length on error, otherwise the length of the payload only, which
	// is the package that is sent out on the local interface.
	return f_lastError != 0 ? 0 : length - AESGCM_HEADER_LENGTH;
}


	// Same as above, but uses tag to authenticate the message.
int AESGCM_DecryptAuth( char* buffer, int length, char** pDecoded )
{
	uint8 initVect[ AESGCM_IV_LENGTH ];

	// Read initialization vector from message (space is re-used for buffer trailing)
	memcpy( initVect, BufferIV( buffer ), LengthIV( ) );

	*pDecoded = (char*)BufferDecOutput( buffer ); // after decoding, we only use the payload

	f_lastError = gcm_auth_decrypt(
		&f_ctxGCM, NetLength( length ),
		initVect, LengthIV( ),
		BufferAAD( buffer ), LengthAAD( ), // additional data and length
		BufferTAG( buffer ), LengthTAG( ), // Tag received
		BufferInput( buffer ),   // input pointer
		(uint8*)*pDecoded );  // output pointer

	return f_lastError != 0 ? 0 : NetLength( length );
}



#endif /* HAVE_SSL */
