/*  
    VTun - Virtual Tunnel over TCP/IP network.

    Copyright (C) 1998-2000  Maxim Krasnyansky <max_mk@yahoo.com>

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    Schrack-Seconet/TU Wien:
    Functions for AES GCM-mode de/encryption. Used in auth.c and lfd_encrypt.c
 */

/*
 * schrack_aes-gcm_crypt; GG
 */ 
#ifndef _SCHRACK_AES_INC_
#define _SCHRACK_AES_INC_

#ifndef __APPLE_CC__
#include <polarssl/entropy.h>
#include <polarssl/ctr_drbg.h>
#include <polarssl/md5.h>
#include <polarssl/gcm.h>
#else
#include <sys/md5.h>
#include <crypto/gcm.h>
#endif  /* __APPLE_CC__ */

typedef unsigned char uint8;
typedef unsigned short uint16;

/* Error codes ****************************************************************/

#define ERR_BASE                             -0x0012  /* First code for error codes in this module */
#define POLARSSL_ERR_AESGCM_GENERIC_ERROR    ERR_BASE - 1  /**< Not meant for actual use. */
#undef ERR_BASE

/* Bit sizes of block and initialization vector *******************************/

#define AES_BLOCKSIZE_BITS 128
#define AES_BLOCK_LENGTH (AES_BLOCKSIZE_BITS>>3)

#define AES_KEYSIZE_BITS AES_BLOCKSIZE_BITS
#define AES_KEY_LENGTH (AES_KEYSIZE_BITS>>3)
   // Note: MD5 is used to hash the AES key from the password in HashKey().
   // If blocksize is to be changed, key generation must be adapted there.

#define AESGCM_BLOCKSIZE_BITS 128
#define AESGCM_BLOCK_LENGTH (AESGCM_BLOCKSIZE_BITS>>3)

#define AESGCM_KEYSIZE_BITS AESGCM_BLOCKSIZE_BITS
#define AESGCM_KEY_LENGTH (AESGCM_KEYSIZE_BITS>>3)

#define AESGCM_IV_BITSIZE 96
#define AESGCM_IV_LENGTH (AESGCM_IV_BITSIZE>>3)
   // IV of 12 bytes/96 bits, the remainder to 128 is used as counter

/* Sizes of fields in AES header **********************************************/

// Additional Authenticated Data fields:
#define AESGCM_AAD_SEQ_LENGTH 2 /* Sequence number for the packet */
#define AESGCM_AAD_LENGTH (AESGCM_AAD_SEQ_LENGTH) /* total AAD length */

#define AESGCM_TAG_LENGTH 8 /* use 8 bytes of the authentication tag */

#define AESGCM_BUFFER_TRAIL 8
   // When decoding into the same buffer, output must be shifted by 8 bytes


// Header is composed of AAD, TAG, and IV.
#define AESGCM_HEADER_LENGTH (AESGCM_AAD_LENGTH + AESGCM_TAG_LENGTH + AESGCM_IV_LENGTH )

#define AESGCM_ENCRYPT GCM_ENCRYPT
#define AESGCM_DECRYPT GCM_DECRYPT     0

/* Generic helper functions ***************************************************/

int AES_GetError( void );


/* Random number generation ***************************************************/

int RAND_bytes( uint8* bytes, int count );


/* AES en/decryption **********************************************************/

typedef aes_context TCtxAES;

typedef struct
{
   uint16 seqNum;
} TAddData;

int AES_InitEncrypt( char* key, int keyLength );
int AES_InitDecrypt( char* key, int keyLength );
void AES_Exit( void );
int AES_EncryptECB( char* aInput, char* aOutput );
int AES_DecryptECB( char* aInput, char* aOutput );


/* AES-GCM en/decryption ******************************************************/

typedef gcm_context TCtxGCM;

int AESGCM_Init( char* key, int keyLength );
void AESGCM_Exit( void );
int AESGCM_Encrypt( char* buffer, int length, char** pOutput );
int AESGCM_Decrypt( char* buffer, int length, char** pDecoded );
int AESGCM_DecryptAuth( char* buffer, int length, char** pDecoded );


#endif // _SCHRACK_AES_INC_
